/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: opendiskfont.c,v 1.3 2015/02/04 16:06:36 cyfm Exp $
 
    Desc: Diskfont function OpenDiskFont()
    Lang: english
*/

#ifndef TURN_OFF_DEBUG
#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG 0
#endif
#  include <aros/debug.h>

#include <graphics/text.h>
#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/graphics.h>
#include <proto/diskfont.h>
#include "diskfont_intern.h"


/*****************************************************************************
 
    NAME */
#include <clib/diskfont_protos.h>

AROS_LH1(struct TextFont *, OpenDiskFont,

         /*  SYNOPSIS */
         AROS_LHA(struct TextAttr *, textAttr, A0),

         /*  LOCATION */
         struct Library *, DiskfontBase, 5, Diskfont)

/*  FUNCTION
		Tries to open the font specified by textAttr. If the font has allready
		been loaded into memory, it will be opened with OpenFont(). Otherwise
		OpenDiskFont() will try to load it from disk.
 
    INPUTS
    	textAttr - Description of the font to load. If the textAttr->ta_Style
    			   FSF_TAGGED bit is set, it will be treated as a struct TTextAttr.
    	
 
    RESULT
    	Pointer to a struct TextFont on success, 0 on failure.
 
    NOTES
 
    EXAMPLE
 
    BUGS
    	Loadseg_AOS() which is internally used to load a font file
    	from disk, does not handle address relocation for 64 bit CPUs
    	yet. Will add a hack to support this.
 
    SEE ALSO
    	AvailFonts()
 
    INTERNALS
 
    HISTORY
	27-11-96    digulla automatically created from
			    diskfont_lib.fd and clib/diskfont_protos.h
 
*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct Library *,DiskfontBase)

	WORD 			match_weight = 0, new_match_weight;
	struct FontHookCommand 	fhc;
	struct Hook 		*hook, *bestmatch_hook = NULL;
	struct TTextAttr 		best_so_far =
	    {
		    0
	    }
	    , *destattr;
	struct TextFont 		*tf = 0;
	APTR    	    	    	bestmatch_userdata = 0;
	UWORD 			idx;

	BOOL 			finished = FALSE;
	BOOL 			matchfound = FALSE;

	D(bug("OpenDiskFont(textAttr=%p)\n", textAttr));
	D(bug("Name %s YSize %ld Style 0x%lx Flags 0x%lx\n",
	      textAttr->ta_Name,
	      textAttr->ta_YSize,
	      textAttr->ta_Style,
	      textAttr->ta_Flags));

	tf=OpenFont(textAttr);
	if (tf)
	{
		D(bug("openfont tf 0x%lx '%s' size %d\n", tf,tf->tf_Message.mn_Node.ln_Name,tf->tf_YSize));
		if (WeighTAMatch((struct TTextAttr *) textAttr, (struct TextAttr *)(((int)&tf->tf_YSize)-4),
		                 ((struct TextFontExtension *)tf->tf_Extension)->tfe_Tags) != MAXFONTMATCHWEIGHT)
		{
			D(bug("weight is not right..search better font\n"));
			CloseFont(tf);
			tf=NULL;
		}
	}

	if (!tf)
	{

		/* PerfectMatch info should be passed to hook */

		fhc.fhc_ReqAttr = (struct TTextAttr*)textAttr;
		destattr = &(fhc.fhc_DestTAttr);

		/* Ask the hooks for matching textattrs */

		for (idx = 1; (idx < NUMFONTHOOKS) && !finished; idx ++)
		{
			ULONG retval;

			hook = &(DFB(DiskfontBase)->hdescr[idx].ahd_Hook);

			D(bug("fonthook[%ld] 0x%lx\n",idx,hook));

			/* Initalize the hook */
			fhc.fhc_Command = FHC_ODF_INIT;

			if ( !CallHookPkt(hook, &fhc, DFB(DiskfontBase) ))
			{
				D(bug("Init failed\n"));
				continue; /* Go on with next hook */
			}

			D(bug("try to match it\n"));

			/* Ask the hook for info on a font */
			fhc.fhc_Command = FHC_ODF_GETMATCHINFO;

			for (;;)
			{
				/* Reset tags field in case the hook doesn't do it */
				destattr->tta_Tags = NULL;

				retval = CallHookPkt(hook, &fhc, DFB(DiskfontBase));
				D(bug("retval 0x%lx\n",retval));

				/* Error or finished scanning ? */
				if (!retval || (retval & FH_SCANFINISHED))
				{
					D(bug("match failed\n"));
					break;
				}

#if 0
				/* WeighTAMatch does not compare the fontnames, so we do it here */
				if ( strcmp(FilePart(textAttr->ta_Name), destattr->tta_Name) != 0 )
				{
					continue;
				}
#endif

				/* How well does this font match ? */
				new_match_weight = WeighTAMatch((struct TTextAttr *) textAttr, (struct TextAttr *)destattr, destattr->tta_Tags);

				D(bug("New matchweight: %d destAttr '%s' size %d\n", new_match_weight,destattr->tta_Name,destattr->tta_YSize));

				/* Better match found ? */

				if (new_match_weight > match_weight)
				{
					matchfound = TRUE;

					match_weight = new_match_weight;

					if (bestmatch_hook && (bestmatch_hook != hook))
					{
						/* Old bestmatch_hook is no longer best matching hook,
						   so we send it the FHC_ODF_CLEANUP, which it initially
						   did not got sent, because of --> [1] */

						struct FontHookCommand cleanup_fhc = fhc;

						cleanup_fhc.fhc_Command  = FHC_ODF_CLEANUP;
						cleanup_fhc.fhc_UserData = bestmatch_userdata;

						D(bug("cleanup font\n"));
						CallHookPkt(bestmatch_hook, &cleanup_fhc, DFB(DiskfontBase) );

					}

					best_so_far = *destattr;

					bestmatch_hook 	   = hook;
					bestmatch_userdata = fhc.fhc_UserData;

					/* Perfect match found ? */
					if (new_match_weight == MAXFONTMATCHWEIGHT)
					{
						finished = TRUE;
						D(bug("\tPerfect match\n"));
						break;
					}

				}

			} /* for (;;) */

			if (bestmatch_hook != hook)
			{
				/* Tell the hook to cleanup */
				fhc.fhc_Command = FHC_ODF_CLEANUP;

				D(bug("cleanup font\n"));
				CallHookPkt(hook, &fhc, DFB(DiskfontBase) );
			}
			else
			{
				/*
				** [1] The hook of the best matching font will get FHC_ODF_CLEANUP
				**     after FHC_ODF_OPENFONT
				*/
			}


		} /* for ( iterate hooktable ) */

		if (matchfound)
		{
			STRPTR savepath;

			D(bug("match found\n"));

			/* Save and replace best_so_far.tta_Name because it contains full path, not just the name! */
			savepath = best_so_far.tta_Name;
			best_so_far.tta_Name = textAttr->ta_Name;

			/* Check if best match already is loaded! */
			if ((tf = OpenFont((struct TextAttr *)&best_so_far)))
			{
				D(bug("check weight of matching font in memory\n"));

				if (WeighTAMatch(&best_so_far, (struct TextAttr *)(((int)&tf->tf_YSize)-4), ((struct TextFontExtension *)tf->tf_Extension)->tfe_Tags) != MAXFONTMATCHWEIGHT)
				{
					D(bug("weight is not right\n"));
					CloseFont(tf);
					tf = NULL;
				}
			}

			/* Restore */
			best_so_far.tta_Name = savepath;

			if (!tf)
			{
				/* Open the font */
				fhc.fhc_Command     	    = FHC_ODF_OPENFONT;
				fhc.fhc_UserData    	    = bestmatch_userdata;
				fhc.fhc_ReqAttr     	    = (struct TTextAttr*)textAttr;
				fhc.fhc_DestTAttr           = best_so_far;

				D(bug("open best match\n"));

				CallHookPkt(bestmatch_hook, &fhc, DFB(DiskfontBase) );
				tf = fhc.fhc_TextFont;
			}

			D(bug("tf 0x%lx '%s' size %d\n",tf,tf->tf_Message.mn_Node.ln_Name,tf->tf_YSize));

			if (tf)
			{
#ifdef __MORPHOS__
				if((tf->tf_YSize != textAttr->ta_YSize) && !(textAttr->ta_Flags & FPF_DESIGNED))
				{
					if(!(tf->tf_Style & FSF_COLORFONT)
					|| (((struct ColorTextFont *)tf)->ctf_Flags == CT_COLORFONT))	/* only scale legacy colorfonts */
					{
						struct DiskFontHeader *dfh;

						if((dfh = (struct DiskFontHeader *)NewScaledDiskFont(tf,textAttr)) != NULL)
						{
							struct DiskfontBase_intern *DiskFontBase=(struct DiskfontBase_intern *)DiskfontBase;
							Forbid();
							AddFont(&dfh->dfh_TF);
								dfh->dfh_TF.tf_Accessors++;
							ADDTAIL(&DiskFontBase->diskfontlist, &dfh->dfh_DF);
							Permit();
							tf = &dfh->dfh_TF;
						}
					}
				}
#endif

				D(bug("extend font\n"));
				if (ExtendFont(tf, NULL))
				{
#warning CHECKME

					struct TextFontExtension *tfe = (struct TextFontExtension *)tf->tf_Extension;


					tfe->tfe_Flags0 |= TE0F_NOREMFONT;
				}
			}


			/* cleanup only here, because of [1] */

			fhc.fhc_Command = FHC_ODF_CLEANUP;

			D(bug("cleanup font\n"));

			CallHookPkt(bestmatch_hook, &fhc, DFB(DiskfontBase) );

		}
	}

	ReturnPtr("OpenDiskFont", struct TextFont *, tf);

	AROS_LIBFUNC_EXIT

} /* OpenDiskFont */
